
class EmptyParameter(Exception):
    """Represents an empty parameter"""
    pass

class EmptyAnnotationError(Exception):
    """"""
    pass

class PlaceHolerKeyError(Exception):
    """"""
    pass

class RedundantParameterError(Exception):
    """"""
    pass

class ParameterTypeError(Exception):
    """"""
    pass

class DatabaseNotFoundError(Exception):
    """"""
    pass


class ResultTypeError(Exception):
    """"""
    pass


class ValueDerivationError(Exception):
    """"""
    pass


class MultipleDerivationError(Exception):
    """"""
    pass


class DependencyTypeError(Exception):
    """"""
    pass


class ResultNotFound(Exception):
    """"""
    pass

